#include "fompi.h"

// this should really be dome by autoconf but I don't have the time right now
#define F77_FUNC_(name,NAME) name ## _

foMPI_Win *Fortran_Windows=NULL;

// this should also be done more carefully (see libnbc's handling) -- no time
void F77_FUNC_(fompi_init,FOMPI_INIT)(int *ierr);
void F77_FUNC_(fompi_init,FOMPI_INIT)(int *ierr) {
	
  *ierr = foMPI_Init( NULL, NULL);
}

void F77_FUNC_(fompi_finalize,FOMPI_FINALIZE)(int *ierr);
void F77_FUNC_(fompi_finalize,FOMPI_FINALIZE)(int *ierr) {
	
  *ierr = foMPI_Finalize( NULL, NULL);
}

void F77_FUNC_(fompi_win_create,FOMPI_WIN_CREATE)(void *base, int *size, int *disp_unit, int *info, int *comm, int *win, int *ierr);
void F77_FUNC_(fompi_win_create,FOMPI_WIN_CREATE)(void *base, int *size, int *disp_unit, int *info, int *icomm, int *win, int *ierr) {
	
  // if we don't have a window table yet ... allocate it
  const int sz = 100; // no more windows that that ... also not nice
  int winidx=0;
  if(Fortran_Windows == NULL) {
    int i;
    Fortran_Windows = (foMPI_Win*)malloc(sizeof(foMPI_Win)*sz);
    for (i=0; i<sz; ++i) Fortran_Windows[i] = NULL;
  } else {
    int i=0;
    for (i=0; i<sz; ++i) if(Fortran_Windows[i] == NULL) break;
    //assert(i!=sz);
    if(i == sz) { MPI_Abort(MPI_COMM_WORLD, 35); }
    winidx=i;
  }

  MPI_Comm comm;
  comm = MPI_Comm_f2c(*icomm);

  // this is also very suboptimal
  *ierr = foMPI_Win_create(base, (MPI_Aint)*size, *disp_unit, MPI_INFO_NULL, comm, &Fortran_Windows[winidx]);
  *win = winidx;
}


void F77_FUNC_(fompi_win_lock,FOMPI_WIN_LOCK)(int *type, int *rank, int *assert, int *win, int *ierr);
void F77_FUNC_(fompi_win_lock,FOMPI_WIN_LOCK)(int *type, int *rank, int *assert, int *win, int *ierr) {
 
  *ierr = foMPI_Win_lock(*type, *rank, *assert, Fortran_Windows[*win]);
 
}

void F77_FUNC_(fompi_win_unlock,FOMPI_WIN_UNLOCK)(int *rank, int *win, int *ierr);
void F77_FUNC_(fompi_win_unlock,FOMPI_WIN_UNLOCK)(int *rank, int *win, int *ierr) {
 
  *ierr = foMPI_Win_unlock(*rank, Fortran_Windows[*win]);
 
}

void F77_FUNC_(fompi_win_fence,FOMPI_WIN_FENCE)(int *assert, int *win, int *ierr);
void F77_FUNC_(fompi_win_fence,FOMPI_WIN_FENCE)(int *assert, int *win, int *ierr) {
 
  *ierr = foMPI_Win_fence(*assert, Fortran_Windows[*win]);
 
}

void F77_FUNC_(fompi_accumulate,FOMPI_ACCUMULATE)(void *origin_addr, int *origin_count, int *origin_datatype, int *target_rank, int *target_disp, int *target_count, int *target_datatype, int *op, int *win, int *ierr);
void F77_FUNC_(fompi_accumulate,FOMPI_ACCUMULATE)(void *origin_addr, int *origin_count, int *origin_datatype, int *target_rank, int *target_disp, int *target_count, int *target_datatype, int *op, int *win, int *ierr){

  MPI_Datatype otype, ttype;
  otype = MPI_Type_f2c(*origin_datatype);
  ttype = MPI_Type_f2c(*target_datatype);
  MPI_Op mop;
  mop = *op;//MPI_Op_f2c(*op); not for now, since we use foMPI prefix operators

  //printf("val: %i, count: %i, target: %i, disp: %i, count: %i: %p, %i\n", *(int*)origin_addr, *origin_count, *target_rank, *target_disp, *target_count);

  *ierr = foMPI_Accumulate(origin_addr, *origin_count, otype, *target_rank, (MPI_Aint)*target_disp, *target_count, ttype, mop, Fortran_Windows[*win]);

  foMPI_Win_flush(*target_rank, Fortran_Windows[*win]);

}

void F77_FUNC_(fompi_put,FOMPI_PUT)(void *origin_addr, int *origin_count, int *origin_datatype, int *target_rank, int *target_disp, int *target_count, int *target_datatype, int *win, int *ierr);
void F77_FUNC_(fompi_put,FOMPI_PUT)(void *origin_addr, int *origin_count, int *origin_datatype, int *target_rank, int *target_disp, int *target_count, int *target_datatype, int *win, int *ierr){

  MPI_Datatype otype, ttype;
  otype = MPI_Type_f2c(*origin_datatype);
  ttype = MPI_Type_f2c(*target_datatype);

  //printf("val: %i, count: %i, target: %i, disp: %i, count: %i\n", *(int*)origin_addr, *origin_count, *target_rank, *target_disp, *target_count);

  *ierr = foMPI_Put(origin_addr, *origin_count, otype, *target_rank, (MPI_Aint)*target_disp, *target_count, ttype, Fortran_Windows[*win]);

}

void F77_FUNC_(fompi_win_free,FOMPI_WIN_FREE)(int *win, int *ierr);
void F77_FUNC_(fompi_win_free,FOMPI_WIN_FREE)(int *win, int *ierr) {

  *ierr = foMPI_Win_free(&Fortran_Windows[*win]);
  Fortran_Windows[*win] = NULL;

}
