#include <assert.h>
//#include <intrinsics.h> 
#include <string.h>
#include "fompi.h"

/* TODO: only does busy waiting, should be doing at least some queueing */
/* TODO: asserts unused */
int foMPI_Win_lock(int lock_type, int rank, int assert, foMPI_Win win) {
#ifdef PAPI
  timing_record( -1 );
#endif
  dmapp_return_t status;
  uint64_t result;
  uint64_t bit_mask = ~(uint64_t)0 << 32;

  foMPI_Win_desc_t* win_ptr;
  dmapp_seg_desc_t* win_ptr_seg;
  foMPI_Win_desc_t* win_ptr_master;
  dmapp_seg_desc_t* win_ptr_seg_master;
 
  assert( (lock_type == foMPI_LOCK_SHARED) || (lock_type == foMPI_LOCK_EXCLUSIVE) );

  if (win->epoch_type == foMPI_WIN_EPOCH_REFENCE) {
    win->epoch_type = foMPI_WIN_EPOCH_NONE;
  }

  /* TODO: is this consistent with the standard */
  assert( win->win_locks[rank] == foMPI_LOCK_NONE );
 
  if ( rank == MPI_PROC_NULL ) {
    return MPI_SUCCESS;
  }
 
  assert((rank >= 0) && (rank < win->commsize));
  dmapp_pe_t target_pe = (dmapp_pe_t) win->group_ranks[rank];

  if ( win->create_flavor == foMPI_WIN_FLAVOR_DYNAMIC ) {
    win_ptr = win->dynamic_list[rank].win_ptr;
    win_ptr_seg = &(win->dynamic_list[rank].win_ptr_seg);
    win_ptr_master = win->dynamic_list[MASTER].win_ptr;
    win_ptr_seg_master = &(win->dynamic_list[MASTER].win_ptr_seg);
  } else {
    win_ptr = win->win_array[rank].win_ptr;
    win_ptr_seg = &(win->win_array[rank].win_ptr_seg);
    win_ptr_master = win->win_array[MASTER].win_ptr;
    win_ptr_seg_master = &(win->win_array[MASTER].win_ptr_seg);
  }

  if (lock_type == foMPI_LOCK_EXCLUSIVE) {
    if ( win->local_exclusive_count == 0) {
      /* check if lock_all mutex is set on the master */
      /* if not, then acquire lock_all mutex */
      /* TODO: howto to count exclusive locks in the system? */
      dmapp_pe_t master_pe = (dmapp_pe_t) win->group_ranks[MASTER];

      do {
        /* check if lock_all mutex is set on the master */
        /* if not, then acquire lock_all mutex */
        /* TODO: are we in trouble, because of int64_t operands in acswap and uint64_t variable */
        status = dmapp_afadd_qw( &result, (void*) &(win_ptr_master->lock_all_mutex), win_ptr_seg_master, master_pe, (int64_t) 1 );
        _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);

        if (!( result & bit_mask )) {
          /* try to acquire exclusive on target */
          status = dmapp_acswap_qw( &result, (void*) &(win_ptr->lock_mutex), win_ptr_seg, target_pe, (int64_t) 0, (int64_t) 1 );
          _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
        
          /* if not acquired back off: release lock_all mutex  */
          if (result == 0) {
            break;
          } else {
            /* TODO: non-blocking? */
            status = dmapp_aadd_qw( (void*) &(win_ptr_master->lock_all_mutex), win_ptr_seg_master, master_pe, (int64_t) -1 );
            _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
          }
        } else {
          /* back off on master lock_all mutex */
          /* TODO: non-blocking? */
          status = dmapp_aadd_qw( (void*) &(win_ptr_master->lock_all_mutex), win_ptr_seg_master, master_pe, (int64_t) -1 );
          _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
        }
      } while (1);
      win->local_exclusive_count++;
    } else {
      /* we already hold the global exclusive lock */
      /* try to acquire exclusive on target */
      do {
        status = dmapp_acswap_qw( &result, (void*) &(win_ptr->lock_mutex), win_ptr_seg, target_pe, (int64_t) 0, (int64_t) 1 );
        _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
      } while ( result != 0 );
      win->local_exclusive_count++;
    }

  } else { /* foMPI_LOCK_SHARED */
    /* propagate read intention for target process */
    /* TODO: non-blocking? */
    status = dmapp_afadd_qw( (void*) &result, (void*) &(win_ptr->lock_mutex), win_ptr_seg, target_pe, (int64_t) 2 );
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
      
    /* check if there is an exclusive lock */
    while ( result & 1) {
      status = dmapp_get( &result, (void*) &(win_ptr->lock_mutex), win_ptr_seg, target_pe, 1, DMAPP_QW );
      _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
    };
  }

  win->win_locks[rank] = lock_type;

#ifdef PAPI
  timing_record( 10 );
#endif
  return MPI_SUCCESS;
}

int foMPI_Win_unlock(int rank, foMPI_Win win) {
#ifdef PAPI
  timing_record( -1 );
#endif
  dmapp_return_t status;
  foMPI_Win_desc_t* win_ptr;
  dmapp_seg_desc_t* win_ptr_seg;
  foMPI_Win_desc_t* win_ptr_master;
  dmapp_seg_desc_t* win_ptr_seg_master;

  if ( rank == MPI_PROC_NULL ) {
    return MPI_SUCCESS;
  }

  assert((rank >= 0) && (rank < win->commsize));
  dmapp_pe_t target_pe = (dmapp_pe_t) win->group_ranks[rank];

  assert( win->win_locks[win->commsize] != foMPI_LOCK_SHARED_ALL );

  if ( win->create_flavor == foMPI_WIN_FLAVOR_DYNAMIC ) {
    win_ptr = win->dynamic_list[rank].win_ptr;
    win_ptr_seg = &(win->dynamic_list[rank].win_ptr_seg);
    win_ptr_master = win->dynamic_list[MASTER].win_ptr;
    win_ptr_seg_master = &(win->dynamic_list[MASTER].win_ptr_seg);
  } else {
    win_ptr = win->win_array[rank].win_ptr;
    win_ptr_seg = &(win->win_array[rank].win_ptr_seg);
    win_ptr_master = win->win_array[MASTER].win_ptr;
    win_ptr_seg_master = &(win->win_array[MASTER].win_ptr_seg);
  }

#ifdef PAPI
  timing_record( 8 );
#endif

  foMPI_Win_flush( rank, win );

  /* release the lock(s) */
  if ( win->win_locks[rank] == foMPI_LOCK_EXCLUSIVE) {
    dmapp_pe_t master_pe = (dmapp_pe_t) win->group_ranks[MASTER];

    /* release lock on target */
    /* TODO: non-blocking? */
    status = dmapp_aadd_qw( (void*) &(win_ptr->lock_mutex), win_ptr_seg, target_pe, (int64_t) -1 );
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
  
    win->local_exclusive_count--;

    if ( win->local_exclusive_count == 0 ) {
      /* release lock on master process */
     status = dmapp_aadd_qw_nbi( (void*) &(win_ptr_master->lock_all_mutex), win_ptr_seg_master, master_pe, (int64_t) -1 );
      _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
    }
 
  } else { /* foMPI_LOCK_SHARED */
    /* release lock on target */
    status = dmapp_aadd_qw_nbi( (void*) &(win_ptr->lock_mutex), win_ptr_seg, target_pe, (int64_t) -2 );
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
  }
  
  win->win_locks[rank] = foMPI_LOCK_NONE;

#ifdef PAPI
  timing_record( 9 );
#endif

  return MPI_SUCCESS;
}

int foMPI_Win_lock_all(int assert, foMPI_Win win) {
#ifdef PAPI
  timing_record( -1 );
#endif

  dmapp_return_t status;
  uint64_t result;
  uint64_t bit_mask = ~(~(uint64_t)0 << 32);

  foMPI_Win_desc_t* win_ptr_master;
  dmapp_seg_desc_t* win_ptr_seg_master;
  dmapp_pe_t master_pe = (dmapp_pe_t) win->group_ranks[MASTER];
 
  if ( win->create_flavor == foMPI_WIN_FLAVOR_DYNAMIC ) {
    win_ptr_master = win->dynamic_list[MASTER].win_ptr;
    win_ptr_seg_master = &(win->dynamic_list[MASTER].win_ptr_seg);
  } else {
    win_ptr_master = win->win_array[MASTER].win_ptr;
    win_ptr_seg_master = &(win->win_array[MASTER].win_ptr_seg);
  }

  if (win->epoch_type == foMPI_WIN_EPOCH_REFENCE) {
    win->epoch_type = foMPI_WIN_EPOCH_NONE;
  }

  status = dmapp_afadd_qw( (void*) &result, (void*) &(win_ptr_master->lock_all_mutex), win_ptr_seg_master, master_pe, (int64_t) 4294967296 ); /* 2^32 = 4294967296 */
  _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);

  while( result & bit_mask ) {
    status = dmapp_get( &result, (void*) &(win_ptr_master->lock_all_mutex), win_ptr_seg_master, master_pe, 1, DMAPP_QW );
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
  }

  win->win_locks[win->commsize] = foMPI_LOCK_SHARED_ALL;

#ifdef PAPI
  timing_record( 16 );
#endif

  return MPI_SUCCESS;
}

int foMPI_Win_unlock_all(foMPI_Win win) {
#ifdef PAPI
  timing_record( -1 );
#endif

  dmapp_return_t status;

  foMPI_Win_desc_t* win_ptr_master;
  dmapp_seg_desc_t* win_ptr_seg_master;

  assert( win->win_locks[win->commsize] == foMPI_LOCK_SHARED_ALL );

  /* target and local completion */
  foMPI_Win_flush_all( win );
  
  /* release lock on master process */
  dmapp_pe_t master_pe = (dmapp_pe_t) win->group_ranks[MASTER];
 
  if ( win->create_flavor == foMPI_WIN_FLAVOR_DYNAMIC ) {
    win_ptr_master = win->dynamic_list[MASTER].win_ptr;
    win_ptr_seg_master = &(win->dynamic_list[MASTER].win_ptr_seg);
  } else {
    win_ptr_master = win->win_array[MASTER].win_ptr;
    win_ptr_seg_master = &(win->win_array[MASTER].win_ptr_seg);
  }
  status = dmapp_aadd_qw_nbi( (void*) &(win_ptr_master->lock_all_mutex), win_ptr_seg_master, master_pe, (int64_t) -4294967296 ); /* 2^32 = 4294967296 */
  _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);

  win->win_locks[win->commsize] = foMPI_LOCK_NONE;

#ifdef PAPI
  timing_record( 17 );
#endif

  return MPI_SUCCESS;
}

/* TODO: write a test for the flush functions */
int foMPI_Win_flush(int rank, foMPI_Win win) {

#ifdef PAPI
  timing_record( -1 );
#endif

  dmapp_return_t status;

  if ( rank == MPI_PROC_NULL ) {
    return MPI_SUCCESS;
  }

  assert((rank >= 0) && (rank < win->commsize));

#ifdef XPMEM
  /* XPMEM completion */
  __sync_synchronize();
#endif

  /* completion */
  if (win->nbi_counter > 0) {
    status = dmapp_gsync_wait();
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
    win->nbi_counter = 0;
  }

#ifdef PAPI
  timing_record( 4 );
#endif

  return MPI_SUCCESS;
}

int foMPI_Win_flush_local(int rank, foMPI_Win win) {
  return foMPI_Win_flush(rank, win);
}

int foMPI_Win_flush_all(foMPI_Win win) {
  return foMPI_Win_flush( 0 /* dummy */, win);
}

int foMPI_Win_flush_local_all(foMPI_Win win) {
  return foMPI_Win_flush( 0 /* dummy */, win);
}

int foMPI_Win_sync(foMPI_Win win) {
  return MPI_SUCCESS;
}
